@@ -16,6 +16,15 @@
 #include "efa-abi.h"
 #include "efa_io_defs.h"
 
+#define EFA_GET(ptr, mask) FIELD_GET(mask##_MASK, *(ptr))
+
+#define EFA_SET(ptr, mask, value)                                              \
+	({                                                                     \
+		typeof(ptr) _ptr = ptr;                                        \
+		*_ptr = (*_ptr & ~(mask##_MASK)) |                             \
+			FIELD_PREP(mask##_MASK, value);                        \
+	})
+
 struct efa_context {
 	struct verbs_context ibvctx;
 	uint32_t cmds_supp_udata_mask;
@@ -57,6 +66,9 @@ struct efa_cq {
 	size_t cqe_size;
 	uint8_t *buf;
 	size_t buf_size;
+	uint32_t *db;
+	uint16_t cc; /* Consumer Counter */
+	uint8_t cmd_sn;
 	uint16_t num_sub_cqs;
 	/* Index of next sub cq idx to poll. This is used to guarantee fairness for sub cqs */
 	uint16_t next_poll_idx;
