@@ -375,7 +375,15 @@ static bool dr_mask_is_flex_parser_0_3_set(struct dr_match_misc4 *misc4)
 		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_2,
 			misc4->prog_sample_field_value_2) ||
 		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_3,
-			misc4->prog_sample_field_value_3));
+			misc4->prog_sample_field_value_3) ||
+		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_4,
+			misc4->prog_sample_field_value_4) ||
+		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_5,
+			misc4->prog_sample_field_value_5) ||
+		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_6,
+			misc4->prog_sample_field_value_6) ||
+		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_7,
+			misc4->prog_sample_field_value_7));
 }
 
 static bool dr_mask_is_flex_parser_id_4_7_set(uint32_t flex_parser_id)
@@ -389,7 +397,11 @@ static bool dr_mask_is_flex_parser_4_7_set(struct dr_match_misc4 *misc4)
 	return (dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_0) ||
 		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_1) ||
 		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_2) ||
-		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_3));
+		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_3) ||
+		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_4) ||
+		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_5) ||
+		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_6) ||
+		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_7));
 }
 
 static bool dr_mask_is_tunnel_header_0_1_set(struct dr_match_misc5 *misc5)
@@ -613,6 +625,16 @@ static int dr_matcher_set_definer_builders(struct mlx5dv_dr_matcher *matcher,
 		idx = 0;
 	}
 
+	if (dmn->info.caps.definer_format_sup & (1ULL << DR_MATCHER_DEFINER_33)) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def33(ste_ctx, &sb[idx++], &mask, false, rx);
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(sb, 0, sizeof(*sb));
+		idx = 0;
+	}
+
 	return ENOTSUP;
 
 done:
@@ -620,6 +642,16 @@ done:
 	return 0;
 }
 
+static bool dr_matcher_is_definer_support_mq(struct dr_matcher_rx_tx *nic_matcher)
+{
+	/* ipv6 needs 2 definers and not supported yet */
+	if (nic_matcher->num_of_builders == 1 &&
+	    nic_matcher->ste_builder->htbl_type == DR_STE_HTBL_TYPE_MATCH)
+		return true;
+
+	return false;
+}
+
 static int dr_matcher_set_large_ste_builders(struct mlx5dv_dr_matcher *matcher,
 					     struct dr_matcher_rx_tx *nic_matcher)
 {
@@ -933,7 +965,7 @@ static int dr_matcher_connect(struct mlx5dv_dr_domain *dmn,
 	}
 	ret = dr_ste_htbl_init_and_postsend(dmn, nic_dmn,
 					    curr_nic_matcher->e_anchor,
-					    &info, info.type == CONNECT_HIT);
+					    &info, info.type == CONNECT_HIT, 0);
 	if (ret)
 		return ret;
 
@@ -942,7 +974,7 @@ static int dr_matcher_connect(struct mlx5dv_dr_domain *dmn,
 	info.miss_icm_addr = curr_nic_matcher->e_anchor->chunk->icm_addr;
 	ret = dr_ste_htbl_init_and_postsend(dmn, nic_dmn,
 					    curr_nic_matcher->s_htbl,
-					    &info, false);
+					    &info, false, 0);
 	if (ret)
 		return ret;
 
@@ -955,7 +987,7 @@ static int dr_matcher_connect(struct mlx5dv_dr_domain *dmn,
 	info.type = CONNECT_HIT;
 	info.hit_next_htbl = curr_nic_matcher->s_htbl;
 	ret = dr_ste_htbl_init_and_postsend(dmn, nic_dmn, prev_htbl,
-					    &info, true);
+					    &info, true, 0);
 	if (ret)
 		return ret;
 
@@ -1159,6 +1191,11 @@ static int dr_matcher_init_root(struct mlx5dv_dr_matcher *matcher,
 	return 0;
 }
 
+static bool dr_matcher_is_fixed_size(struct mlx5dv_dr_matcher *matcher)
+{
+	return (matcher->rx.fixed_size || matcher->tx.fixed_size);
+}
+
 static int dr_matcher_init(struct mlx5dv_dr_matcher *matcher,
 			   struct mlx5dv_flow_match_parameters *mask)
 {
@@ -1204,9 +1241,106 @@ static int dr_matcher_init(struct mlx5dv_dr_matcher *matcher,
 		return errno;
 	}
 
+	/* Drain QP to resolve possible race between new multi QP rules
+	 * and matcher hash table initial creation.
+	 */
+	if (dr_matcher_is_fixed_size(matcher))
+		dr_send_ring_force_drain(dmn);
+
 	return ret;
 }
 
+static int
+dr_matcher_set_nic_matcher_layout(struct mlx5dv_dr_matcher *matcher,
+				  struct dr_matcher_rx_tx *nic_matcher,
+				  struct mlx5dv_dr_matcher_layout *matcher_layout)
+{
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	int ret = 0;
+
+	if (!dr_matcher_is_definer_support_mq(nic_matcher)) {
+		dr_dbg(dmn, "not supported not a definer\n");
+		errno = ENOTSUP;
+		return ENOTSUP;
+	}
+
+	dr_domain_lock(dmn);
+
+	if (matcher_layout->flags & MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE) {
+		/* if needed set dmn->info.max_log_sw_icm_sz and pool max_log_chunk_sz */
+		dr_domain_set_max_ste_icm_size(dmn, matcher_layout->log_num_of_rules_hint);
+
+		ret = dr_rule_rehash_matcher_s_anchor(matcher, nic_matcher,
+						      matcher_layout->log_num_of_rules_hint);
+		if (ret) {
+			dr_dbg(dmn, "failed rehash with log-size: %d\n",
+			       matcher_layout->log_num_of_rules_hint);
+			goto out;
+		}
+	}
+
+	if (matcher_layout->flags & MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE) {
+		nic_matcher->fixed_size = false;
+	} else {
+		nic_matcher->fixed_size = true;
+		dmn->info.use_mqs = true;
+	}
+
+	dr_send_ring_force_drain(dmn);
+out:
+	dr_domain_unlock(dmn);
+	return ret;
+}
+
+int mlx5dv_dr_matcher_set_layout(struct mlx5dv_dr_matcher *matcher,
+				 struct mlx5dv_dr_matcher_layout *matcher_layout)
+{
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	int ret = 0;
+
+	if (dr_is_root_table(matcher->tbl)) {
+		dr_dbg(dmn, "Not supported in root table\n");
+		errno = ENOTSUP;
+		return ENOTSUP;
+	}
+	if (!check_comp_mask(matcher_layout->flags,
+			 MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE |
+			 MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE)) {
+		dr_dbg(dmn, "Not supported flags 0x%x\n", matcher_layout->flags);
+		errno = ENOTSUP;
+		return ENOTSUP;
+	}
+
+	if ((matcher_layout->flags & MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE) &&
+	    !dr_domain_is_support_ste_icm_size(dmn, matcher_layout->log_num_of_rules_hint)) {
+		dr_dbg(dmn, "the size is too big: %d\n",
+		       matcher_layout->log_num_of_rules_hint);
+		errno = ENOTSUP;
+		return ENOTSUP;
+	}
+
+	if (dmn->type == MLX5DV_DR_DOMAIN_TYPE_NIC_RX ||
+	    dmn->type == MLX5DV_DR_DOMAIN_TYPE_FDB) {
+		ret = dr_matcher_set_nic_matcher_layout(matcher,
+							&matcher->rx,
+							matcher_layout);
+	}
+	if (!ret && (dmn->type == MLX5DV_DR_DOMAIN_TYPE_NIC_TX ||
+		     dmn->type == MLX5DV_DR_DOMAIN_TYPE_FDB)) {
+		ret = dr_matcher_set_nic_matcher_layout(matcher,
+							&matcher->tx,
+							matcher_layout);
+	}
+
+	if (ret) {
+		dr_dbg(dmn, "failed nic (%d) rehash with log-size: %d\n",
+		       dmn->type, matcher_layout->log_num_of_rules_hint);
+		return ret;
+	}
+
+	return 0;
+}
+
 struct mlx5dv_dr_matcher *
 mlx5dv_dr_matcher_create(struct mlx5dv_dr_table *tbl,
 			 uint16_t priority,
@@ -1282,7 +1416,7 @@ static int dr_matcher_disconnect(struct mlx5dv_dr_domain *dmn,
 	}
 
 	return dr_ste_htbl_init_and_postsend(dmn, nic_dmn, prev_anchor,
-					     &info, true);
+					     &info, true, 0);
 }
 
 static int dr_matcher_remove_from_tbl(struct mlx5dv_dr_matcher *matcher)
