@@ -34,7 +34,8 @@
 #include <ccan/minmax.h>
 #include "mlx5dv_dr.h"
 
-#define DR_RULE_MAX_STE_CHAIN (DR_RULE_MAX_STES + DR_ACTION_MAX_STES)
+/* +1 for the cross GVMI STE */
+#define DR_RULE_MAX_STE_CHAIN (DR_RULE_MAX_STES + DR_ACTION_MAX_STES + 1)
 
 static int dr_rule_append_to_miss_list(struct dr_ste_ctx *ste_ctx,
 				       struct dr_ste *new_last_ste,
@@ -99,7 +100,8 @@ static struct dr_ste
 static struct dr_ste *dr_rule_create_collision_entry(struct mlx5dv_dr_matcher *matcher,
 						     struct dr_matcher_rx_tx *nic_matcher,
 						     uint8_t *hw_ste,
-						     struct dr_ste *orig_ste)
+						     struct dr_ste *orig_ste,
+						     uint8_t send_ring_idx)
 {
 	struct dr_ste *ste;
 
@@ -117,7 +119,7 @@ static struct dr_ste *dr_rule_create_collision_entry(struct mlx5dv_dr_matcher *m
 
 	/* Next table */
 	if (dr_ste_create_next_htbl(matcher, nic_matcher, ste, hw_ste,
-				    DR_CHUNK_SIZE_1)) {
+				    DR_CHUNK_SIZE_1, send_ring_idx)) {
 		dr_dbg(matcher->tbl->dmn, "Failed allocating table\n");
 		goto free_tbl;
 	}
@@ -130,7 +132,8 @@ free_tbl:
 }
 
 static int dr_rule_handle_one_ste_in_update_list(struct dr_ste_send_info *ste_info,
-						 struct mlx5dv_dr_domain *dmn)
+						 struct mlx5dv_dr_domain *dmn,
+						 uint8_t send_ring_idx)
 {
 	int ret;
 
@@ -145,7 +148,8 @@ static int dr_rule_handle_one_ste_in_update_list(struct dr_ste_send_info *ste_in
 		memcpy(ste_info->ste->hw_ste, ste_info->data, ste_info->ste->size);
 
 	ret = dr_send_postsend_ste(dmn, ste_info->ste, ste_info->data,
-				   ste_info->size, ste_info->offset);
+				   ste_info->size, ste_info->offset,
+				   send_ring_idx);
 	if (ret)
 		goto out;
 
@@ -154,9 +158,10 @@ out:
 	return ret;
 }
 
-static int dr_rule_send_update_list(struct list_head *send_ste_list,
-				    struct mlx5dv_dr_domain *dmn,
-				    bool is_reverse)
+int dr_rule_send_update_list(struct list_head *send_ste_list,
+			     struct mlx5dv_dr_domain *dmn,
+			     bool is_reverse,
+			     uint8_t send_ring_idx)
 {
 	struct dr_ste_send_info *ste_info, *tmp_ste_info;
 	int ret;
@@ -165,7 +170,8 @@ static int dr_rule_send_update_list(struct list_head *send_ste_list,
 		list_for_each_rev_safe(send_ste_list, ste_info, tmp_ste_info,
 				       send_list) {
 			ret = dr_rule_handle_one_ste_in_update_list(ste_info,
-								    dmn);
+								    dmn,
+								    send_ring_idx);
 			if (ret)
 				return ret;
 		}
@@ -173,7 +179,8 @@ static int dr_rule_send_update_list(struct list_head *send_ste_list,
 		list_for_each_safe(send_ste_list, ste_info, tmp_ste_info,
 				   send_list) {
 			ret = dr_rule_handle_one_ste_in_update_list(ste_info,
-								    dmn);
+								    dmn,
+								    send_ring_idx);
 			if (ret)
 				return ret;
 		}
@@ -308,7 +315,8 @@ static struct dr_ste *dr_rule_rehash_copy_ste(struct mlx5dv_dr_matcher *matcher,
 		if (!ste_info) {
 			dr_dbg(matcher->tbl->dmn, "Failed allocating ste_info\n");
 			errno = ENOMEM;
-			goto err_exit;
+			dr_htbl_put(new_ste->htbl);
+			return NULL;
 		}
 		dr_send_fill_and_append_ste_send_info(new_ste, DR_STE_SIZE, 0,
 						      hw_ste, ste_info,
@@ -318,10 +326,6 @@ static struct dr_ste *dr_rule_rehash_copy_ste(struct mlx5dv_dr_matcher *matcher,
 	dr_rule_rehash_copy_ste_ctrl(matcher, nic_matcher, cur_ste, new_ste);
 
 	return new_ste;
-
-err_exit:
-	dr_ste_free(new_ste, matcher, nic_matcher);
-	return NULL;
 }
 
 static int dr_rule_rehash_copy_miss_list(struct mlx5dv_dr_matcher *matcher,
@@ -383,16 +387,15 @@ clean_copy:
 	return err;
 }
 
-static struct dr_ste_htbl *dr_rule_rehash_htbl(struct mlx5dv_dr_rule *rule,
-					       struct dr_rule_rx_tx *nic_rule,
-					       struct dr_ste_htbl *cur_htbl,
-					       uint8_t ste_location,
-					       struct list_head *update_list,
-					       enum dr_icm_chunk_size new_size)
+static struct dr_ste_htbl *dr_rule_rehash_htbl_common(struct mlx5dv_dr_matcher *matcher,
+						      struct dr_matcher_rx_tx *nic_matcher,
+						      struct dr_ste_htbl *cur_htbl,
+						      uint8_t ste_location,
+						      struct list_head *update_list,
+						      enum dr_icm_chunk_size new_size,
+						      uint8_t lock_index)
 {
-	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
 	struct dr_domain_rx_tx *nic_dmn = nic_matcher->nic_tbl->nic_dmn;
-	struct mlx5dv_dr_matcher *matcher = rule->matcher;
 	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
 	struct dr_ste_send_info *del_ste_info, *tmp_ste_info;
 	uint8_t formated_ste[DR_STE_SIZE] = {};
@@ -443,7 +446,7 @@ static struct dr_ste_htbl *dr_rule_rehash_htbl(struct mlx5dv_dr_rule *rule,
 	if (new_htbl->type == DR_STE_HTBL_TYPE_LEGACY)
 		mask = nic_matcher->ste_builder[ste_location - 1].bit_mask;
 
-	if (dr_send_postsend_htbl(dmn, new_htbl, formated_ste, mask)) {
+	if (dr_send_postsend_htbl(dmn, new_htbl, formated_ste, mask, lock_index)) {
 		dr_dbg(dmn, "Failed writing table to HW\n");
 		goto free_new_htbl;
 	}
@@ -453,7 +456,8 @@ static struct dr_ste_htbl *dr_rule_rehash_htbl(struct mlx5dv_dr_rule *rule,
 	 * in order to have the origin data written before the miss address of
 	 * collision entries, if exists.
 	 */
-	if (dr_rule_send_update_list(&rehash_table_send_list, dmn, false)) {
+	if (dr_rule_send_update_list(&rehash_table_send_list, dmn, false,
+				     lock_index)) {
 		dr_dbg(dmn, "Failed updating table to HW\n");
 		goto free_ste_list;
 	}
@@ -507,6 +511,60 @@ free_ste_info:
 	return NULL;
 }
 
+static struct dr_ste_htbl *dr_rule_rehash_htbl(struct mlx5dv_dr_rule *rule,
+					       struct dr_rule_rx_tx *nic_rule,
+					       struct dr_ste_htbl *cur_htbl,
+					       uint8_t ste_location,
+					       struct list_head *update_list,
+					       enum dr_icm_chunk_size new_size)
+{
+	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
+	struct mlx5dv_dr_matcher *matcher = rule->matcher;
+
+	return dr_rule_rehash_htbl_common(matcher, nic_matcher, cur_htbl,
+					  ste_location, update_list, new_size,
+					  nic_rule->lock_index);
+}
+
+int dr_rule_rehash_matcher_s_anchor(struct mlx5dv_dr_matcher *matcher,
+				    struct dr_matcher_rx_tx *nic_matcher,
+				    enum dr_icm_chunk_size new_size)
+{
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	struct dr_ste_htbl *new_htbl;
+	LIST_HEAD(update_list);
+	int ret;
+
+	if (nic_matcher->s_htbl->chunk_size == new_size) {
+		dr_dbg(dmn, "both are with the same size, nothing to do\n");
+		return 0;
+	}
+
+	new_htbl = dr_rule_rehash_htbl_common(matcher, nic_matcher,
+					      nic_matcher->s_htbl,
+					      1, &update_list, new_size, 0);
+	if (!new_htbl) {
+		dr_dbg(dmn, "Failed creating new matcher s_anchor\n");
+		goto err_out;
+	}
+
+	ret = dr_rule_send_update_list(&update_list, dmn, true, 0);
+	if (ret) {
+		dr_dbg(dmn, "Failed sending ste!\n");
+		goto free_new_htbl;
+	}
+
+	dr_ste_htbl_free(nic_matcher->s_htbl);
+	nic_matcher->s_htbl = new_htbl;
+
+	return 0;
+
+free_new_htbl:
+	dr_ste_htbl_free(new_htbl);
+err_out:
+	return ENOTSUP;
+}
+
 static struct dr_ste_htbl *dr_rule_rehash(struct mlx5dv_dr_rule *rule,
 					  struct dr_rule_rx_tx *nic_rule,
 					  struct dr_ste_htbl *cur_htbl,
@@ -527,12 +585,13 @@ static struct dr_ste_htbl *dr_rule_rehash(struct mlx5dv_dr_rule *rule,
 }
 
 static struct dr_ste *dr_rule_handle_collision(struct mlx5dv_dr_matcher *matcher,
-					       struct dr_matcher_rx_tx *nic_matcher,
+					       struct dr_rule_rx_tx *nic_rule,
 					       struct dr_ste *ste,
 					       uint8_t *hw_ste,
 					       struct list_head *miss_list,
 					       struct list_head *send_list)
 {
+	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
 	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
 	struct dr_ste_ctx *ste_ctx = dmn->ste_ctx;
 	struct dr_ste_send_info *ste_info;
@@ -545,7 +604,8 @@ static struct dr_ste *dr_rule_handle_collision(struct mlx5dv_dr_matcher *matcher
 		return NULL;
 	}
 
-	new_ste = dr_rule_create_collision_entry(matcher, nic_matcher, hw_ste, ste);
+	new_ste = dr_rule_create_collision_entry(matcher, nic_matcher, hw_ste,
+						 ste, nic_rule->lock_index);
 	if (!new_ste) {
 		dr_dbg(dmn, "Failed creating collision entry\n");
 		goto free_send_info;
@@ -650,14 +710,37 @@ void dr_rule_get_reverse_rule_members(struct dr_ste **ste_arr,
 static void dr_rule_clean_rule_members(struct mlx5dv_dr_rule *rule,
 				       struct dr_rule_rx_tx *nic_rule)
 {
-	struct dr_ste *ste_arr[DR_RULE_MAX_STES + DR_ACTION_MAX_STES];
+	struct dr_ste *ste_arr[DR_RULE_MAX_STES + DR_ACTION_MAX_STES +
+			       DR_ACTION_ASO_CROSS_GVMI_STES];
 	struct dr_ste *curr_ste = nic_rule->last_rule_ste;
 	int i;
 
 	dr_rule_get_reverse_rule_members(ste_arr, curr_ste, &i);
 
 	while (i--)
-		dr_ste_put(ste_arr[i], rule->matcher, nic_rule->nic_matcher);
+		dr_ste_put(ste_arr[i], rule, nic_rule);
+}
+
+static void dr_rule_clean_cross_dmn_rule_members(struct mlx5dv_dr_rule *rule,
+						 struct dr_rule_rx_tx *nic_rule,
+						 struct list_head *send_ste_list,
+						 struct mlx5dv_dr_action *cross_dmn_action)
+{
+	struct dr_aso_cross_dmn_arrays *cross_dmn_arrays =
+		(struct dr_aso_cross_dmn_arrays *) cross_dmn_action->aso.devx_obj->priv;
+	struct dr_ste_send_info *ste_info, *tmp_ste_info;
+
+	dr_rule_clean_rule_members(rule, nic_rule);
+	/* Clean all ste_info's */
+	list_for_each_safe(send_ste_list, ste_info, tmp_ste_info, send_list) {
+		list_del(&ste_info->send_list);
+		free(ste_info);
+	}
+
+	if (atomic_load(&cross_dmn_arrays->rule_htbl[cross_dmn_action->aso.offset]->ste_arr->refcount) > 1) {
+		atomic_fetch_sub(&cross_dmn_arrays->rule_htbl[cross_dmn_action->aso.offset]->ste_arr->refcount, 1);
+		atomic_fetch_sub(&cross_dmn_arrays->action_htbl[cross_dmn_action->aso.offset]->ste_arr->refcount, 1);
+	}
 }
 
 static uint16_t dr_get_bits_per_mask(uint16_t byte_mask)
@@ -672,6 +755,106 @@ static uint16_t dr_get_bits_per_mask(uint16_t byte_mask)
 	return bits;
 }
 
+static int dr_rule_handle_cross_action_stes(struct mlx5dv_dr_rule *rule,
+					    struct dr_rule_rx_tx *nic_rule,
+					    struct list_head *send_ste_list,
+					    struct dr_ste *last_ste,
+					    uint8_t *hw_ste_arr,
+					    uint32_t new_hw_ste_arr_sz,
+					    uint32_t cross_dmn_loc,
+					    struct mlx5dv_dr_action *cross_dmn_action)
+{
+	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
+	struct dr_ste_send_info *ste_info_arr[DR_ACTION_MAX_STES + 1];
+	uint8_t num_of_builders = nic_matcher->num_of_builders;
+	struct mlx5dv_dr_matcher *matcher = rule->matcher;
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	uint8_t *curr_hw_ste, *prev_hw_ste;
+	struct dr_ste *action_ste, *cross_dmn_action_ste,
+		      *cross_dmn_rule_ste;
+	bool is_ste_for_cross_dmn;
+	int i, k;
+
+	for (i = num_of_builders, k = 0; i < new_hw_ste_arr_sz; i++, k++) {
+		curr_hw_ste = hw_ste_arr + i * DR_STE_SIZE;
+
+		prev_hw_ste = (i == 0) ? curr_hw_ste :
+			hw_ste_arr + (i - 1) * DR_STE_SIZE;
+
+		is_ste_for_cross_dmn = (cross_dmn_loc == k);
+
+		if (!is_ste_for_cross_dmn) {
+			action_ste = dr_rule_create_collision_htbl(matcher,
+								   nic_matcher,
+								   curr_hw_ste);
+			if (!action_ste)
+				return errno;
+
+			dr_ste_get(action_ste);
+
+			action_ste->htbl->pointing_ste = last_ste;
+			last_ste->next_htbl = action_ste->htbl;
+			last_ste = action_ste;
+
+			/* While free ste we go over the miss list, so add this ste to the list */
+			list_add_tail(dr_ste_get_miss_list(action_ste),
+				      &action_ste->miss_list_node);
+
+			/* Point current ste to the new action */
+			dr_ste_set_hit_addr_by_next_htbl(dmn->ste_ctx,
+							 prev_hw_ste,
+							 action_ste->htbl);
+
+			dr_rule_set_last_member(nic_rule, action_ste, true);
+		} else {
+			struct dr_aso_cross_dmn_arrays *cross_dmn_arrays = (struct dr_aso_cross_dmn_arrays *) cross_dmn_action->aso.devx_obj->priv;
+
+			cross_dmn_action_ste = cross_dmn_arrays->action_htbl[cross_dmn_action->aso.offset]->ste_arr;
+			cross_dmn_rule_ste = cross_dmn_arrays->rule_htbl[cross_dmn_action->aso.offset]->ste_arr;
+			/* Connect last ste to cross_dmn_action_ste */
+			cross_dmn_action_ste->htbl->pointing_ste = last_ste;
+			last_ste->next_htbl = cross_dmn_action_ste->htbl;
+			dr_ste_set_hit_addr_by_next_htbl(dmn->ste_ctx,
+							 prev_hw_ste,
+							 cross_dmn_action_ste->htbl);
+			dr_ste_set_hit_gvmi(dmn->ste_ctx,
+					    prev_hw_ste,
+					    cross_dmn_action->aso.dmn->info.caps.gvmi);
+			/* Point rule STE as last member */
+			dr_rule_set_last_member(nic_rule, cross_dmn_rule_ste, true);
+			/* Point rule STE as last STE */
+			last_ste = cross_dmn_rule_ste;
+		}
+
+		ste_info_arr[k] = calloc(1, sizeof(struct dr_ste_send_info));
+		if (!ste_info_arr[k]) {
+			dr_dbg(dmn, "Failed allocate ste_info, k: %d\n", k);
+			errno = ENOMEM;
+			return errno;
+		}
+
+		if (!is_ste_for_cross_dmn) {
+			dr_send_fill_and_append_ste_send_info(action_ste, DR_STE_SIZE, 0,
+							      curr_hw_ste,
+							      ste_info_arr[k],
+							      send_ste_list, false);
+		} else {
+			memcpy(cross_dmn_rule_ste->hw_ste,
+			       curr_hw_ste,
+			       DR_STE_SIZE_REDUCED);
+			dr_send_fill_and_append_ste_send_info(cross_dmn_rule_ste,
+							      DR_STE_SIZE, 0,
+							      curr_hw_ste,
+							      ste_info_arr[k],
+							      send_ste_list, false);
+		}
+	}
+
+	last_ste->next_htbl = NULL;
+
+	return 0;
+}
+
 static bool dr_rule_need_enlarge_hash(struct dr_ste_htbl *htbl,
 				      struct mlx5dv_dr_domain *dmn,
 				      struct dr_domain_rx_tx *nic_dmn)
@@ -697,12 +880,12 @@ static bool dr_rule_need_enlarge_hash(struct dr_ste_htbl *htbl,
 	return false;
 }
 
-static int dr_rule_handle_action_stes(struct mlx5dv_dr_rule *rule,
-				      struct dr_rule_rx_tx *nic_rule,
-				      struct list_head *send_ste_list,
-				      struct dr_ste *last_ste,
-				      uint8_t *hw_ste_arr,
-				      uint32_t new_hw_ste_arr_sz)
+static int dr_rule_handle_regular_action_stes(struct mlx5dv_dr_rule *rule,
+					      struct dr_rule_rx_tx *nic_rule,
+					      struct list_head *send_ste_list,
+					      struct dr_ste *last_ste,
+					      uint8_t *hw_ste_arr,
+					      uint32_t new_hw_ste_arr_sz)
 {
 	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
 	struct dr_ste_send_info *ste_info_arr[DR_ACTION_MAX_STES];
@@ -769,12 +952,36 @@ static int dr_rule_handle_action_stes(struct mlx5dv_dr_rule *rule,
 	return 0;
 
 err_exit:
-	dr_ste_put(action_ste, matcher, nic_matcher);
+	dr_ste_put(action_ste, rule, nic_rule);
 	return ret;
 }
 
+static int dr_rule_handle_action_stes(struct mlx5dv_dr_rule *rule,
+				      struct dr_rule_rx_tx *nic_rule,
+				      struct list_head *send_ste_list,
+				      struct dr_ste *last_ste,
+				      uint8_t *hw_ste_arr,
+				      uint32_t new_hw_ste_arr_sz,
+				      struct cross_dmn_params *cross_dmn_p)
+{
+	if (cross_dmn_p->cross_dmn_loc != -1)
+		return dr_rule_handle_cross_action_stes(rule, nic_rule,
+							send_ste_list,
+							last_ste,
+							hw_ste_arr,
+							new_hw_ste_arr_sz,
+							cross_dmn_p->cross_dmn_loc,
+							cross_dmn_p->cross_dmn_action);
+
+	return dr_rule_handle_regular_action_stes(rule, nic_rule,
+						  send_ste_list,
+						  last_ste,
+						  hw_ste_arr,
+						  new_hw_ste_arr_sz);
+}
+
 static int dr_rule_handle_empty_entry(struct mlx5dv_dr_matcher *matcher,
-				      struct dr_matcher_rx_tx *nic_matcher,
+				      struct dr_rule_rx_tx *nic_rule,
 				      struct dr_ste_htbl *cur_htbl,
 				      struct dr_ste *ste,
 				      uint8_t ste_location,
@@ -782,6 +989,7 @@ static int dr_rule_handle_empty_entry(struct mlx5dv_dr_matcher *matcher,
 				      struct list_head *miss_list,
 				      struct list_head *send_list)
 {
+	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
 	struct dr_ste_ctx *ste_ctx = matcher->tbl->dmn->ste_ctx;
 	struct dr_ste_send_info *ste_info;
 
@@ -806,7 +1014,8 @@ static int dr_rule_handle_empty_entry(struct mlx5dv_dr_matcher *matcher,
 				    nic_matcher,
 				    ste,
 				    hw_ste,
-				    DR_CHUNK_SIZE_1)) {
+				    DR_CHUNK_SIZE_1,
+				    nic_rule->lock_index)) {
 		dr_dbg(matcher->tbl->dmn, "Failed allocating table\n");
 		goto clean_ste_info;
 	}
@@ -843,7 +1052,7 @@ static struct dr_ste *dr_rule_handle_ste_branch(struct mlx5dv_dr_rule *rule,
 	struct dr_ste_htbl *new_htbl;
 	struct list_head *miss_list;
 	struct dr_ste *matched_ste;
-	bool skip_rehash = false;
+	bool skip_rehash = nic_matcher->fixed_size;
 	struct dr_ste *ste;
 	int index;
 
@@ -853,7 +1062,7 @@ again:
 	ste = &cur_htbl->ste_arr[index];
 
 	if (dr_ste_is_not_used(ste)) {
-		if (dr_rule_handle_empty_entry(matcher, nic_matcher, cur_htbl,
+		if (dr_rule_handle_empty_entry(matcher, nic_rule, cur_htbl,
 					       ste, ste_location,
 					       hw_ste, miss_list,
 					       send_ste_list))
@@ -903,7 +1112,7 @@ again:
 		} else {
 			/* Hash table index in use, add another collision (miss) */
 			ste = dr_rule_handle_collision(matcher,
-						       nic_matcher,
+						       nic_rule,
 						       ste,
 						       hw_ste,
 						       miss_list,
@@ -1030,9 +1239,9 @@ static bool dr_rule_verify(struct mlx5dv_dr_matcher *matcher,
 static int dr_rule_destroy_rule_nic(struct mlx5dv_dr_rule *rule,
 				    struct dr_rule_rx_tx *nic_rule)
 {
-	dr_domain_nic_lock(nic_rule->nic_matcher->nic_tbl->nic_dmn);
+	dr_rule_lock(nic_rule, NULL);
 	dr_rule_clean_rule_members(rule, nic_rule);
-	dr_domain_nic_unlock(nic_rule->nic_matcher->nic_tbl->nic_dmn);
+	dr_rule_unlock(nic_rule);
 	return 0;
 }
 
@@ -1047,9 +1256,9 @@ static int dr_rule_destroy_rule(struct mlx5dv_dr_rule *rule)
 {
 	struct mlx5dv_dr_domain *dmn = rule->matcher->tbl->dmn;
 
-	dr_domain_lock(dmn);
+	pthread_spin_lock(&dmn->debug_lock);
 	list_del(&rule->rule_list);
-	dr_domain_unlock(dmn);
+	pthread_spin_unlock(&dmn->debug_lock);
 
 	switch (dmn->type) {
 	case MLX5DV_DR_DOMAIN_TYPE_NIC_RX:
@@ -1086,12 +1295,12 @@ static int dr_rule_destroy_rule_root(struct mlx5dv_dr_rule *rule)
 	return 0;
 }
 
-static int dr_rule_skip(enum mlx5dv_dr_domain_type domain,
+static int dr_rule_skip(struct mlx5dv_dr_domain *dmn,
 			enum dr_domain_nic_type nic_type,
 			struct dr_match_param *mask,
 			struct dr_match_param *value)
 {
-	if (domain == MLX5DV_DR_DOMAIN_TYPE_FDB) {
+	if (dmn->type == MLX5DV_DR_DOMAIN_TYPE_FDB) {
 		if (mask->misc.source_port) {
 			if (nic_type == DR_DOMAIN_NIC_TYPE_RX)
 				if (value->misc.source_port != WIRE_PORT)
@@ -1101,6 +1310,30 @@ static int dr_rule_skip(enum mlx5dv_dr_domain_type domain,
 				if (value->misc.source_port == WIRE_PORT)
 					return 1;
 		}
+
+		/* Metadata C can be used to describe the source vport */
+		if (mask->misc2.metadata_reg_c_0) {
+			struct dr_devx_vport_cap *wire;
+			uint32_t vport_metadata_c;
+
+			wire = &dmn->info.caps.vports.wire;
+
+			/* No correlation between wire mask and regc0 mask */
+			if (!(wire->metadata_c_mask &
+			      mask->misc2.metadata_reg_c_0))
+				return 0;
+
+			vport_metadata_c = value->misc2.metadata_reg_c_0
+					   & wire->metadata_c_mask;
+
+			if (nic_type == DR_DOMAIN_NIC_TYPE_RX)
+				if (vport_metadata_c != wire->metadata_c)
+					return 1;
+
+			if (nic_type == DR_DOMAIN_NIC_TYPE_TX)
+				if (vport_metadata_c == wire->metadata_c)
+					return 1;
+		}
 	}
 
 	return 0;
@@ -1121,12 +1354,15 @@ dr_rule_create_rule_nic(struct mlx5dv_dr_rule *rule,
 	struct dr_ste_send_info *ste_info, *tmp_ste_info;
 	struct dr_ste_htbl *htbl = NULL;
 	struct dr_ste_htbl *cur_htbl;
-	uint32_t new_hw_ste_arr_sz;
+	uint32_t new_hw_ste_arr_sz = 0;
+	struct cross_dmn_params cross_dmn_p = {};
 	LIST_HEAD(send_ste_list);
 	struct dr_ste *ste = NULL; /* Fix compilation warning */
 	int ret, i;
 
-	if (dr_rule_skip(dmn->type, nic_dmn->type, &matcher->mask, param))
+	cross_dmn_p.cross_dmn_loc = -1;
+
+	if (dr_rule_skip(dmn, nic_dmn->type, &matcher->mask, param))
 		return 0;
 
 	/* Set the tag values inside the ste array */
@@ -1137,11 +1373,12 @@ dr_rule_create_rule_nic(struct mlx5dv_dr_rule *rule,
 	/* Set the actions values/addresses inside the ste array */
 	ret = dr_actions_build_ste_arr(matcher, nic_matcher, actions,
 				       num_actions, hw_ste_arr,
-				       &new_hw_ste_arr_sz);
+				       &new_hw_ste_arr_sz,
+				       &cross_dmn_p);
 	if (ret)
 		return ret;
 
-	dr_domain_nic_lock(nic_rule->nic_matcher->nic_tbl->nic_dmn);
+	dr_rule_lock(nic_rule, hw_ste_arr);
 
 	cur_htbl = nic_matcher->s_htbl;
 
@@ -1173,13 +1410,14 @@ dr_rule_create_rule_nic(struct mlx5dv_dr_rule *rule,
 	}
 
 	/* Connect actions */
-	ret = dr_rule_handle_action_stes(rule, nic_rule, &send_ste_list,
-					 ste, hw_ste_arr, new_hw_ste_arr_sz);
+	ret = dr_rule_handle_action_stes(rule, nic_rule, &send_ste_list, ste,
+					 hw_ste_arr, new_hw_ste_arr_sz,
+					 &cross_dmn_p);
 	if (ret) {
 		dr_dbg(dmn, "Failed apply actions\n");
 		goto free_rule;
 	}
-	ret = dr_rule_send_update_list(&send_ste_list, dmn, true);
+	ret = dr_rule_send_update_list(&send_ste_list, dmn, true, nic_rule->lock_index);
 	if (ret) {
 		dr_dbg(dmn, "Failed sending ste!\n");
 		goto free_rule;
@@ -1191,14 +1429,20 @@ dr_rule_create_rule_nic(struct mlx5dv_dr_rule *rule,
 	goto out_unlock;
 
 free_rule:
-	dr_rule_clean_rule_members(rule, nic_rule);
-	/* Clean all ste_info's */
-	list_for_each_safe(&send_ste_list, ste_info, tmp_ste_info, send_list) {
-		list_del(&ste_info->send_list);
-		free(ste_info);
+	if (cross_dmn_p.cross_dmn_action) {
+		dr_rule_clean_cross_dmn_rule_members(rule, nic_rule,
+						     &send_ste_list,
+						     cross_dmn_p.cross_dmn_action);
+	} else {
+		dr_rule_clean_rule_members(rule, nic_rule);
+		/* Clean all ste_info's */
+		list_for_each_safe(&send_ste_list, ste_info, tmp_ste_info, send_list) {
+			list_del(&ste_info->send_list);
+			free(ste_info);
+		}
 	}
 out_unlock:
-	dr_domain_nic_unlock(nic_rule->nic_matcher->nic_tbl->nic_dmn);
+	dr_rule_unlock(nic_rule);
 	return ret;
 }
 
@@ -1288,9 +1532,9 @@ dr_rule_create_rule(struct mlx5dv_dr_matcher *matcher,
 	if (ret)
 		goto remove_action_members;
 
-	dr_domain_lock(dmn);
+	pthread_spin_lock(&dmn->debug_lock);
 	list_add_tail(&matcher->rule_list, &rule->rule_list);
-	dr_domain_unlock(dmn);
+	pthread_spin_unlock(&dmn->debug_lock);
 
 	return rule;
 
@@ -1341,11 +1585,11 @@ dr_rule_create_rule_root(struct mlx5dv_dr_matcher *matcher,
 	if (ret)
 		goto free_attr_aux;
 
-	rule->flow = __mlx5dv_create_flow(matcher->dv_matcher,
-					  value,
-					  num_actions,
-					  attr,
-					  attr_aux);
+	rule->flow = _mlx5dv_create_flow(matcher->dv_matcher,
+					 value,
+					 num_actions,
+					 attr,
+					 attr_aux);
 	if (!rule->flow)
 		goto remove_action_members;
 
