@@ -43,7 +43,12 @@
 
 #define DR_RULE_MAX_STES	20
 #define DR_ACTION_MAX_STES	7
+#define DR_ACTION_ASO_CROSS_GVMI_STES 2
+/* Use up to 14 send rings. This number provided the best performance */
+#define DR_MAX_SEND_RINGS	14
+#define NUM_OF_LOCKS		DR_MAX_SEND_RINGS
 #define WIRE_PORT		0xFFFF
+#define ECPF_PORT		0xFFFE
 #define DR_STE_SVLAN		0x1
 #define DR_STE_CVLAN		0x2
 #define CVLAN_ETHERTYPE	0x8100
@@ -51,6 +56,7 @@
 #define NUM_OF_FLEX_PARSERS	8
 #define DR_STE_MAX_FLEX_0_ID	3
 #define DR_STE_MAX_FLEX_1_ID	7
+#define DR_VPORTS_BUCKETS	256
 
 #define dr_dbg(dmn, arg...) dr_dbg_ctx((dmn)->ctx, ##arg)
 
@@ -146,7 +152,8 @@ enum dr_matcher_definer {
 	DR_MATCHER_DEFINER_24	= 24,
 	DR_MATCHER_DEFINER_25	= 25,
 	DR_MATCHER_DEFINER_26	= 26,
-	DR_MATCHER_DEFINER_28   = 28
+	DR_MATCHER_DEFINER_28   = 28,
+	DR_MATCHER_DEFINER_33   = 33,
 };
 
 enum dr_action_type {
@@ -313,6 +320,8 @@ void dr_ste_set_hit_addr_by_next_htbl(struct dr_ste_ctx *ste_ctx,
 				      struct dr_ste_htbl *next_htbl);
 void dr_ste_set_hit_addr(struct dr_ste_ctx *ste_ctx, uint8_t *hw_ste_p,
 			 uint64_t icm_addr, uint32_t ht_size);
+void dr_ste_set_hit_gvmi(struct dr_ste_ctx *ste_ctx, uint8_t *hw_ste_p,
+			 uint16_t gvmi);
 void dr_ste_set_bit_mask(uint8_t *hw_ste_p, struct dr_ste_build *sb);
 bool dr_ste_is_last_in_rule(struct dr_matcher_rx_tx *nic_matcher,
 			    uint8_t ste_location);
@@ -331,7 +340,13 @@ static inline int dr_ste_tag_sz(struct dr_ste *ste)
 
 #define MAX_VLANS 2
 
+struct dr_aso_cross_dmn_arrays {
+	struct dr_ste_htbl **action_htbl;
+	struct dr_ste_htbl **rule_htbl;
+};
+
 struct dr_action_aso {
+	struct mlx5dv_dr_domain *dmn;
 	struct mlx5dv_devx_obj *devx_obj;
 	uint32_t offset;
 	uint8_t dest_reg_id;
@@ -367,6 +382,12 @@ struct dr_ste_actions_attr {
 		uint32_t	headers[MAX_VLANS];
 	} vlans;
 	struct dr_action_aso *aso;
+	uint32_t aso_ste_loc;
+};
+
+struct cross_dmn_params {
+	uint32_t cross_dmn_loc;
+	struct mlx5dv_dr_action *cross_dmn_action;
 };
 
 void dr_ste_set_actions_rx(struct dr_ste_ctx *ste_ctx,
@@ -402,6 +423,11 @@ int dr_ste_set_action_decap_l3_list(struct dr_ste_ctx *ste_ctx,
 				   void *data, uint32_t data_sz,
 				   uint8_t *hw_action, uint32_t hw_action_sz,
 				   uint16_t *used_hw_action_num);
+void dr_ste_v1_set_aso_ct(uint8_t *d_action,
+			  uint32_t object_id,
+			  uint32_t offset,
+			  uint8_t dest_reg_id,
+			  bool direction);
 const struct dr_ste_action_modify_field *
 dr_ste_conv_modify_hdr_sw_field(struct dr_ste_ctx *ste_ctx,
 				struct dr_devx_caps *caps,
@@ -409,14 +435,14 @@ dr_ste_conv_modify_hdr_sw_field(struct dr_ste_ctx *ste_ctx,
 
 struct dr_ste_ctx *dr_ste_get_ctx(uint8_t version);
 void dr_ste_free(struct dr_ste *ste,
-		 struct mlx5dv_dr_matcher *matcher,
-		 struct dr_matcher_rx_tx *nic_matcher);
+		 struct mlx5dv_dr_rule *rule,
+		 struct dr_rule_rx_tx *nic_rule);
 static inline void dr_ste_put(struct dr_ste *ste,
-			      struct mlx5dv_dr_matcher *matcher,
-			      struct dr_matcher_rx_tx *nic_matcher)
+			      struct mlx5dv_dr_rule *rule,
+			      struct dr_rule_rx_tx *nic_rule)
 {
 	if (atomic_fetch_sub(&ste->refcount, 1) == 1)
-		dr_ste_free(ste, matcher, nic_matcher);
+		dr_ste_free(ste, rule, nic_rule);
 }
 
 /* initial as 0, increased only when ste appears in a new rule */
@@ -435,7 +461,8 @@ int dr_ste_create_next_htbl(struct mlx5dv_dr_matcher *matcher,
 			    struct dr_matcher_rx_tx *nic_matcher,
 			    struct dr_ste *ste,
 			    uint8_t *cur_hw_ste,
-			    enum dr_icm_chunk_size log_table_size);
+			    enum dr_icm_chunk_size log_table_size,
+			    uint8_t send_ring_idx);
 
 /* STE build functions */
 int dr_ste_build_pre_check(struct mlx5dv_dr_domain *dmn,
@@ -604,6 +631,10 @@ int dr_ste_build_def28(struct dr_ste_ctx *ste_ctx,
 		       struct dr_ste_build *sb,
 		       struct dr_match_param *mask,
 		       bool inner, bool rx);
+int dr_ste_build_def33(struct dr_ste_ctx *ste_ctx,
+		       struct dr_ste_build *sb,
+		       struct dr_match_param *mask,
+		       bool inner, bool rx);
 void dr_ste_build_empty_always_hit(struct dr_ste_build *sb, bool rx);
 
 /* Actions utils */
@@ -612,7 +643,8 @@ int dr_actions_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
 			     struct mlx5dv_dr_action *actions[],
 			     uint32_t num_actions,
 			     uint8_t *ste_arr,
-			     uint32_t *new_hw_ste_arr_sz);
+			     uint32_t *new_hw_ste_arr_sz,
+			     struct cross_dmn_params *cross_dmn_p);
 int dr_actions_build_attr(struct mlx5dv_dr_matcher *matcher,
 			  struct mlx5dv_dr_action *actions[],
 			  size_t num_actions,
@@ -747,6 +779,14 @@ struct dr_match_misc4 {
 	uint32_t prog_sample_field_id_2;
 	uint32_t prog_sample_field_value_3;
 	uint32_t prog_sample_field_id_3;
+	uint32_t prog_sample_field_value_4;
+	uint32_t prog_sample_field_id_4;
+	uint32_t prog_sample_field_value_5;
+	uint32_t prog_sample_field_id_5;
+	uint32_t prog_sample_field_value_6;
+	uint32_t prog_sample_field_id_6;
+	uint32_t prog_sample_field_value_7;
+	uint32_t prog_sample_field_id_7;
 };
 
 struct dr_match_misc5 {
@@ -785,9 +825,15 @@ struct dr_esw_caps {
 };
 
 struct dr_devx_vport_cap {
-	uint16_t gvmi;
+	uint16_t vport_gvmi;
+	uint16_t vhca_gvmi;
 	uint64_t icm_address_rx;
 	uint64_t icm_address_tx;
+	uint16_t num;
+	uint32_t metadata_c;
+	uint32_t metadata_c_mask;
+	/* locate vports table */
+	struct dr_devx_vport_cap *next;
 };
 
 struct dr_devx_roce_cap {
@@ -797,7 +843,27 @@ struct dr_devx_roce_cap {
 	uint8_t qp_ts_format;
 };
 
+struct dr_vports_table {
+	struct dr_devx_vport_cap *buckets[DR_VPORTS_BUCKETS];
+};
+
+struct dr_devx_vports {
+	/* E-Switch manager */
+	struct dr_devx_vport_cap	esw_mngr;
+	/* Uplink */
+	struct dr_devx_vport_cap	wire;
+	/* PF + VFS + SF */
+	struct dr_vports_table		*vports;
+	/* IB ports to vport + other_vports */
+	struct dr_devx_vport_cap	**ib_ports;
+	/* Number of vports PF + VFS + SFS + WIRE */
+	uint32_t			num_ports;
+	/* Protect vport query and add*/
+	pthread_spinlock_t		lock;
+};
+
 struct dr_devx_caps {
+	struct mlx5dv_dr_domain		*dmn;
 	uint16_t			gvmi;
 	uint64_t			nic_rx_drop_address;
 	uint64_t			nic_tx_drop_address;
@@ -831,11 +897,11 @@ struct dr_devx_caps {
 	bool				rx_sw_owner_v2;
 	bool				tx_sw_owner_v2;
 	bool				fdb_sw_owner_v2;
-	uint32_t			num_vports;
-	struct dr_devx_vport_cap	*vports_caps;
 	struct dr_devx_roce_cap		roce_caps;
 	uint64_t			definer_format_sup;
 	bool				prio_tag_required;
+	bool				is_ecpf;
+	struct dr_devx_vports		vports;
 };
 
 struct dr_devx_flow_table_attr {
@@ -903,19 +969,20 @@ struct dr_domain_rx_tx {
 	uint64_t		default_icm_addr;
 	enum dr_domain_nic_type	type;
 	/* protect rx/tx domain */
-	pthread_spinlock_t	lock;
+	pthread_spinlock_t	locks[NUM_OF_LOCKS];
 };
 
 struct dr_domain_info {
 	bool			supp_sw_steering;
 	uint32_t		max_inline_size;
-	uint32_t		max_send_wr;
 	uint32_t		max_log_sw_icm_sz;
 	uint32_t		max_log_action_icm_sz;
+	uint32_t		max_send_size;
 	struct dr_domain_rx_tx	rx;
 	struct dr_domain_rx_tx	tx;
 	struct ibv_device_attr_ex attr;
 	struct dr_devx_caps	caps;
+	bool			use_mqs;
 };
 
 enum dr_domain_flags {
@@ -932,20 +999,57 @@ struct mlx5dv_dr_domain {
 	atomic_int			refcount;
 	struct dr_icm_pool		*ste_icm_pool;
 	struct dr_icm_pool		*action_icm_pool;
-	struct dr_send_ring		*send_ring;
+	struct dr_send_ring		*send_ring[DR_MAX_SEND_RINGS];
 	struct dr_domain_info		info;
 	struct list_head		tbl_list;
 	uint32_t			flags;
+	/* protect debug lists of all tracked objects */
+	pthread_spinlock_t		debug_lock;
 };
 
+static inline int dr_domain_nic_lock_init(struct dr_domain_rx_tx *nic_dmn)
+{
+	int ret;
+	int i;
+
+	for (i = 0; i < NUM_OF_LOCKS; i++) {
+		ret = pthread_spin_init(&nic_dmn->locks[i], PTHREAD_PROCESS_PRIVATE);
+		if (ret) {
+			errno = ret;
+			goto destroy_locks;
+		}
+	}
+	return 0;
+
+destroy_locks:
+	while (i--)
+		pthread_spin_destroy(&nic_dmn->locks[i]);
+
+	return ret;
+}
+
+static inline void dr_domain_nic_lock_uninit(struct dr_domain_rx_tx *nic_dmn)
+{
+	int i;
+
+	for (i = 0; i < NUM_OF_LOCKS; i++)
+		pthread_spin_destroy(&nic_dmn->locks[i]);
+}
+
 static inline void dr_domain_nic_lock(struct dr_domain_rx_tx *nic_dmn)
 {
-	pthread_spin_lock(&nic_dmn->lock);
+	int i;
+
+	for (i = 0; i < NUM_OF_LOCKS; i++)
+		pthread_spin_lock(&nic_dmn->locks[i]);
 }
 
 static inline void dr_domain_nic_unlock(struct dr_domain_rx_tx *nic_dmn)
 {
-	pthread_spin_unlock(&nic_dmn->lock);
+	int i;
+
+	for (i = 0; i < NUM_OF_LOCKS; i++)
+		pthread_spin_unlock(&nic_dmn->locks[i]);
 }
 
 static inline void dr_domain_lock(struct mlx5dv_dr_domain *dmn)
@@ -984,6 +1088,7 @@ struct dr_matcher_rx_tx {
 	uint8_t				num_of_builders;
 	uint64_t			default_icm_addr;
 	struct dr_table_rx_tx		*nic_tbl;
+	bool				fixed_size;
 };
 
 struct mlx5dv_dr_matcher {
@@ -1088,7 +1193,6 @@ struct mlx5dv_dr_action {
 		struct {
 			struct mlx5dv_dr_domain		*dmn;
 			struct dr_devx_vport_cap	*caps;
-			uint32_t			num;
 		} vport;
 		struct {
 			uint32_t	vlan_hdr;
@@ -1127,6 +1231,7 @@ struct dr_htbl_connect_info {
 struct dr_rule_rx_tx {
 	struct dr_matcher_rx_tx		*nic_matcher;
 	struct dr_ste			*last_rule_ste;
+	uint8_t				lock_index;
 };
 
 struct mlx5dv_dr_rule {
@@ -1143,6 +1248,36 @@ struct mlx5dv_dr_rule {
 	uint16_t		num_actions;
 };
 
+static inline void
+dr_rule_lock(struct dr_rule_rx_tx *nic_rule, uint8_t *hw_ste)
+{
+	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
+	struct dr_domain_rx_tx *nic_dmn = nic_matcher->nic_tbl->nic_dmn;
+	uint32_t index;
+
+	if (nic_matcher->fixed_size) {
+		if (hw_ste) {
+			index = dr_ste_calc_hash_index(hw_ste, nic_matcher->s_htbl);
+			nic_rule->lock_index = index % NUM_OF_LOCKS;
+		}
+		pthread_spin_lock(&nic_dmn->locks[nic_rule->lock_index]);
+	} else {
+		pthread_spin_lock(&nic_dmn->locks[0]);
+	}
+}
+
+static inline void
+dr_rule_unlock(struct dr_rule_rx_tx *nic_rule)
+{
+	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
+	struct dr_domain_rx_tx *nic_dmn = nic_matcher->nic_tbl->nic_dmn;
+
+	if (nic_matcher->fixed_size)
+		pthread_spin_unlock(&nic_dmn->locks[nic_rule->lock_index]);
+	else
+		pthread_spin_unlock(&nic_dmn->locks[0]);
+}
+
 void dr_rule_set_last_member(struct dr_rule_rx_tx *nic_rule,
 			     struct dr_ste *ste,
 			     bool force);
@@ -1151,6 +1286,11 @@ void dr_rule_get_reverse_rule_members(struct dr_ste **ste_arr,
 				      struct dr_ste *curr_ste,
 				      int *num_of_stes);
 
+int dr_rule_send_update_list(struct list_head *send_ste_list,
+			     struct mlx5dv_dr_domain *dmn,
+			     bool is_reverse,
+			     uint8_t send_ring_idx);
+
 struct dr_icm_chunk {
 	struct dr_icm_buddy_mem *buddy_mem;
 	struct list_node	chunk_list;
@@ -1199,6 +1339,9 @@ dr_icm_pool_chunk_size_to_byte(enum dr_icm_chunk_size chunk_size,
 	return entry_size * num_of_entries;
 }
 
+void dr_icm_pool_set_pool_max_log_chunk_sz(struct dr_icm_pool *pool,
+					   enum dr_icm_chunk_size max_log_chunk_sz);
+
 static inline int
 dr_ste_htbl_increase_threshold(struct dr_ste_htbl *htbl)
 {
@@ -1219,18 +1362,6 @@ dr_ste_htbl_may_grow(struct dr_ste_htbl *htbl)
 	return true;
 }
 
-static inline struct dr_devx_vport_cap
-*dr_get_vport_cap(struct dr_devx_caps *caps, uint32_t vport)
-{
-	if (!caps->vports_caps ||
-	    (vport >= caps->num_vports && vport != WIRE_PORT)) {
-		errno = EINVAL;
-		return NULL;
-	}
-
-	return &caps->vports_caps[vport == WIRE_PORT ? caps->num_vports : vport];
-}
-
 /* internal API functions */
 int dr_devx_query_device(struct ibv_context *ctx, struct dr_devx_caps *caps);
 int dr_devx_query_esw_vport_context(struct ibv_context *ctx,
@@ -1340,6 +1471,14 @@ static inline bool dr_is_root_table(struct mlx5dv_dr_table *tbl)
 	return tbl->level == 0;
 }
 
+bool dr_domain_is_support_ste_icm_size(struct mlx5dv_dr_domain *dmn,
+				       uint32_t req_log_icm_sz);
+bool dr_domain_set_max_ste_icm_size(struct mlx5dv_dr_domain *dmn,
+				    uint32_t req_log_icm_sz);
+int dr_rule_rehash_matcher_s_anchor(struct mlx5dv_dr_matcher *matcher,
+				    struct dr_matcher_rx_tx *nic_matcher,
+				    enum dr_icm_chunk_size new_size);
+
 struct dr_icm_pool *dr_icm_pool_create(struct mlx5dv_dr_domain *dmn,
 				       enum dr_icm_type icm_type);
 void dr_icm_pool_destroy(struct dr_icm_pool *pool);
@@ -1354,7 +1493,8 @@ int dr_ste_htbl_init_and_postsend(struct mlx5dv_dr_domain *dmn,
 				  struct dr_domain_rx_tx *nic_dmn,
 				  struct dr_ste_htbl *htbl,
 				  struct dr_htbl_connect_info *connect_info,
-				  bool update_hw_ste);
+				  bool update_hw_ste,
+				  uint8_t send_ring_idx);
 void dr_ste_set_formated_ste(struct dr_ste_ctx *ste_ctx,
 			     uint16_t gvmi,
 			     enum dr_domain_nic_type nic_type,
@@ -1408,7 +1548,6 @@ struct dr_cq {
 };
 
 #define MAX_SEND_CQE		64
-#define MIN_READ_SYNC		64
 
 struct dr_send_ring {
 	struct dr_cq		cq;
@@ -1418,31 +1557,32 @@ struct dr_send_ring {
 	uint32_t		pending_wqe;
 	/* Signal request per this trash hold value */
 	uint16_t		signal_th;
-	/* Each post_send_size less than max_post_send_size */
-	uint32_t		max_post_send_size;
+	uint32_t                max_inline_size;
 	/* manage the send queue */
 	uint32_t		tx_head;
 	/* protect QP/CQ operations */
 	pthread_spinlock_t	lock;
 	void			*buf;
 	uint32_t		buf_size;
-	struct ibv_wc		wc[MAX_SEND_CQE];
-	uint8_t			sync_buff[MIN_READ_SYNC];
+	void			*sync_buff;
 	struct ibv_mr		*sync_mr;
 };
 
 int dr_send_ring_alloc(struct mlx5dv_dr_domain *dmn);
-void dr_send_ring_free(struct dr_send_ring *send_ring);
+void dr_send_ring_free(struct mlx5dv_dr_domain *dmn);
 int dr_send_ring_force_drain(struct mlx5dv_dr_domain *dmn);
 bool dr_send_allow_fl(struct dr_devx_caps *caps);
 int dr_send_postsend_ste(struct mlx5dv_dr_domain *dmn, struct dr_ste *ste,
-			 uint8_t *data, uint16_t size, uint16_t offset);
+			 uint8_t *data, uint16_t size, uint16_t offset,
+			 uint8_t ring_idx);
 int dr_send_postsend_htbl(struct mlx5dv_dr_domain *dmn, struct dr_ste_htbl *htbl,
-			  uint8_t *formated_ste, uint8_t *mask);
+			  uint8_t *formated_ste, uint8_t *mask,
+			  uint8_t send_ring_idx);
 int dr_send_postsend_formated_htbl(struct mlx5dv_dr_domain *dmn,
 				   struct dr_ste_htbl *htbl,
 				   uint8_t *ste_init_data,
-				   bool update_hw_ste);
+				   bool update_hw_ste,
+				   uint8_t send_ring_idx);
 int dr_send_postsend_action(struct mlx5dv_dr_domain *dmn,
 			    struct mlx5dv_dr_action *action);
 /* buddy functions & structure */
@@ -1474,4 +1614,14 @@ int dr_buddy_init(struct dr_icm_buddy_mem *buddy, uint32_t max_order);
 void dr_buddy_cleanup(struct dr_icm_buddy_mem *buddy);
 int dr_buddy_alloc_mem(struct dr_icm_buddy_mem *buddy, int order);
 void dr_buddy_free_mem(struct dr_icm_buddy_mem *buddy, uint32_t seg, int order);
+
+void dr_vports_table_add_wire(struct dr_devx_vports *vports);
+void dr_vports_table_del_wire(struct dr_devx_vports *vports);
+struct dr_devx_vport_cap *dr_vports_table_get_vport_cap(struct dr_devx_caps *caps,
+							uint16_t vport);
+struct dr_devx_vport_cap *dr_vports_table_get_ib_port_cap(struct dr_devx_caps *caps,
+							  uint32_t ib_port);
+struct dr_vports_table *dr_vports_table_create(struct mlx5dv_dr_domain *dmn);
+void dr_vports_table_destroy(struct dr_vports_table *vports_tbl);
+
 #endif
