@@ -3,6 +3,7 @@
 """
 Provide some useful helper function for pyverbs rdmacm' tests.
 """
+import sys
 from tests.utils import validate, poll_cq, get_send_elements, get_recv_wr
 from tests.base_rdmacm import AsyncCMResources, SyncCMResources
 from pyverbs.cmid import CMEvent, AddrInfo, JoinMCAttrEx
@@ -15,6 +16,8 @@ import errno
 
 GRH_SIZE = 40
 MULTICAST_QPN = 0xffffff
+REJECT_MSG = 'connection rejected'
+
 
 class CMConnection(abc.ABC):
     """
@@ -202,6 +205,9 @@ class CMConnection(abc.ABC):
         if expected_event and expected_event != cm_event.event_type:
             raise PyverbsError('Expected this event: {}, got this event: {}'.
                                 format(expected_event, cm_event.event_str()))
+        if expected_event == ce.RDMA_CM_EVENT_REJECTED:
+            assert cm_event.private_data == REJECT_MSG, \
+                f'CM event data ({cm_event.private_data}) is different than the expected ({REJECT_MSG})'
         cm_event.ack_cm_event()
 
     @abc.abstractmethod
@@ -218,7 +224,8 @@ class CMAsyncConnection(CMConnection):
     Implement RDMACM connection management for asynchronous CMIDs. It includes
     connection establishment, disconnection and other methods such as traffic.
     """
-    def __init__(self, ip_addr, syncer=None, notifier=None, passive=False, **kwargs):
+    def __init__(self, ip_addr, syncer=None, notifier=None, passive=False,
+                 num_conns=1, qp_timeout=-1, reject_conn=False, **kwargs):
         """
         Init the CMConnection and then init the AsyncCMResources.
         :param ip_addr: IP address to use.
@@ -226,11 +233,25 @@ class CMAsyncConnection(CMConnection):
         :param notifier: Queue object to pass objects between the connection
                          sides.
         :param passive: Indicate if it's a passive side.
+        :param num_conns: Number of connections.
+        :param qp_timeout: Value of the QP timeout.
+        :param reject_conn: True if the server will reject the connection.
         :param kwargs: Arguments used to initialize the CM resources. For more
                        info please check CMResources.
         """
         super(CMAsyncConnection, self).__init__(syncer=syncer, notifier=notifier)
+        self.num_conns = num_conns
+        self.create_cm_res(ip_addr, passive=passive, **kwargs)
+        self.qp_timeout = qp_timeout
+        self.reject_conn = reject_conn
+
+    def create_cm_res(self, ip_addr, passive, **kwargs):
         self.cm_res = AsyncCMResources(addr=ip_addr, passive=passive, **kwargs)
+        if passive:
+            self.cm_res.create_cmid()
+        else:
+            for i in range(self.num_conns):
+                self.cm_res.create_cmid(i)
 
     def join_to_multicast(self, mc_addr=None, src_addr=None, extended=False):
         """
@@ -272,37 +293,60 @@ class CMAsyncConnection(CMConnection):
         if self.cm_res.passive:
             self.cm_res.cmid.bind_addr(self.cm_res.ai)
             self.cm_res.cmid.listen()
-            self.syncer.wait()
-            self.event_handler(expected_event=ce.RDMA_CM_EVENT_CONNECT_REQUEST)
-            self.cm_res.create_qp()
-            if self.cm_res.with_ext_qp:
-                self.set_cmids_qp_ece(self.cm_res.passive)
-                self.cm_res.modify_ext_qp_to_rts()
-                self.set_cmid_ece(self.cm_res.passive)
-            self.cm_res.child_id.accept(self.cm_res.create_conn_param())
-            if self.cm_res.port_space == ce.RDMA_PS_TCP:
-                self.event_handler(expected_event=ce.RDMA_CM_EVENT_ESTABLISHED)
-        else:
-            self.cm_res.cmid.resolve_addr(self.cm_res.ai)
-            self.event_handler(expected_event=ce.RDMA_CM_EVENT_ADDR_RESOLVED)
-            self.syncer.wait()
-            self.cm_res.cmid.resolve_route()
-            self.event_handler(expected_event=ce.RDMA_CM_EVENT_ROUTE_RESOLVED)
-            self.cm_res.create_qp()
-            if self.cm_res.with_ext_qp:
-                self.set_cmid_ece(self.cm_res.passive)
-            self.cm_res.cmid.connect(self.cm_res.create_conn_param())
-            if self.cm_res.with_ext_qp:
-                self.event_handler(expected_event=\
-                    ce.RDMA_CM_EVENT_CONNECT_RESPONSE)
-                self.set_cmids_qp_ece(self.cm_res.passive)
-                self.cm_res.modify_ext_qp_to_rts()
-                self.cm_res.cmid.establish()
+        for conn_idx in range(self.num_conns):
+            if self.cm_res.passive:
+                self.syncer.wait()
+                self.event_handler(expected_event=ce.RDMA_CM_EVENT_CONNECT_REQUEST)
+                self.cm_res.create_qp(conn_idx=conn_idx)
+                if self.qp_timeout >= 0:
+                    self.set_qp_timeout(self.cm_res.child_ids[conn_idx], self.qp_timeout)
+                if self.cm_res.with_ext_qp:
+                    self.set_cmids_qp_ece(self.cm_res.passive)
+                    self.cm_res.modify_ext_qp_to_rts(conn_idx=conn_idx)
+                    self.set_cmid_ece(self.cm_res.passive)
+                child_id = self.cm_res.child_ids[conn_idx]
+                if self.reject_conn:
+                    child_id.reject(REJECT_MSG, sys.getsizeof(REJECT_MSG))
+                    return
+                child_id.accept(self.cm_res.create_conn_param(conn_idx=conn_idx))
+                if self.qp_timeout >= 0:
+                    attr, _ = child_id.query_qp(e.IBV_QP_TIMEOUT)
+                    assert self.qp_timeout == attr.timeout
+                if self.cm_res.port_space == ce.RDMA_PS_TCP:
+                    self.event_handler(expected_event=ce.RDMA_CM_EVENT_ESTABLISHED)
             else:
-                self.event_handler(expected_event=ce.RDMA_CM_EVENT_ESTABLISHED)
+                cmid = self.cm_res.cmids[conn_idx]
+                cmid.resolve_addr(self.cm_res.ai)
+                self.event_handler(expected_event=ce.RDMA_CM_EVENT_ADDR_RESOLVED)
+                self.syncer.wait()
+                cmid.resolve_route()
+                self.event_handler(expected_event=ce.RDMA_CM_EVENT_ROUTE_RESOLVED)
+                self.cm_res.create_qp(conn_idx=conn_idx)
+                if self.qp_timeout >= 0:
+                    self.set_qp_timeout(self.cm_res.cmid, self.qp_timeout)
+                if self.cm_res.with_ext_qp:
+                    self.set_cmid_ece(self.cm_res.passive)
+                cmid.connect(self.cm_res.create_conn_param(conn_idx=conn_idx))
+                if self.cm_res.with_ext_qp:
+                    self.event_handler(expected_event=\
+                        ce.RDMA_CM_EVENT_CONNECT_RESPONSE)
+                    self.set_cmids_qp_ece(self.cm_res.passive)
+                    self.cm_res.modify_ext_qp_to_rts(conn_idx=conn_idx)
+                    cmid.establish()
+                else:
+                    if self.reject_conn:
+                        self.event_handler(expected_event=ce.RDMA_CM_EVENT_REJECTED)
+                        return
+                    self.event_handler(expected_event=ce.RDMA_CM_EVENT_ESTABLISHED)
+                if self.qp_timeout >= 0:
+                    attr, _ = self.cm_res.cmid.query_qp(e.IBV_QP_TIMEOUT)
+                    assert self.qp_timeout == attr.timeout
         self.cm_res.create_mr()
         self.sync_qp_numbers()
 
+    def set_qp_timeout(self, cm_id, ack_timeout):
+        cm_id.set_option(ce.RDMA_OPTION_ID, ce.RDMA_OPTION_ID_ACK_TIMEOUT, ack_timeout, 1)
+
     def sync_qp_numbers(self):
         """
         Sync the QP numbers of the connections sides.
@@ -324,10 +368,12 @@ class CMAsyncConnection(CMConnection):
         """
         if self.cm_res.port_space == ce.RDMA_PS_TCP:
             if self.cm_res.passive:
-                self.cm_res.child_id.disconnect()
+                for child_id in self.cm_res.child_ids.values():
+                    child_id.disconnect()
             else:
                 self.event_handler(expected_event=ce.RDMA_CM_EVENT_DISCONNECTED)
-                self.cm_res.cmid.disconnect()
+                for cmid in self.cm_res.cmids.values():
+                    cmid.disconnect()
 
     def set_cmid_ece(self, passive):
         """
@@ -374,7 +420,11 @@ class CMSyncConnection(CMConnection):
                        info please check CMResources.
         """
         super(CMSyncConnection, self).__init__(syncer=syncer, notifier=notifier)
+        self.create_cm_res(ip_addr, passive=passive, **kwargs)
+
+    def create_cm_res(self, ip_addr, passive, **kwargs):
         self.cm_res = SyncCMResources(addr=ip_addr, passive=passive, **kwargs)
+        self.cm_res.create_cmid()
 
     def establish_connection(self):
         """
